/*
 * Copyright (C) 2008-2010 <sub20@xdimax.net>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.0 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/*
*-------------------------------------------------------------------
* Description: SUB-20 command line application
*-------------------------------------------------------------------
* Version: $Id: sub_upd.c,v 1.6 2010/07/24 21:34:20 avr32 Exp $
*-------------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <libusb.h>
#include <unistd.h>
#include "cmd_pars.h"
#include "sub_upd.h"
#include "usbctrl.h"


/*
*-------------------------------------------------------------------------------
* Constants
*-------------------------------------------------------------------------------
*/

#define SUB_VID		0x04d8
#define SUB_PID		0xffc3

#define MTUSB_VID	0x0abf
#define MTUSB_PID	0x3370

#define WAIT_FOR_DEVICE_TIMEOUT		10

#define WRITE_BUFFER_SIZE		(1 * 1024)


#define min(a,b)	(a)<(b)?(a):(b)

/*
*-------------------------------------------------------------------------------
* Types
*-------------------------------------------------------------------------------
*/

typedef struct upd_fw_context_s
{
	char *buf;
	int size;
	int status ;
	int done ;
}upd_fw_context_t;

typedef struct find_device_context_s
{
	uint16_t sn;
	uint16_t vid;
	uint16_t pid;
	upd_fw_context_t *upd_fw_context;
}find_device_context_t;


/*
*-------------------------------------------------------------------------------
* Global Variables
*-------------------------------------------------------------------------------
*/
libusb_context* usb_context;
int do_not_ask = 0;
/*
*-------------------------------------------------------------------------------
* Local Function Definition
*-------------------------------------------------------------------------------
*/

int update_firmware(char *filename);

/*
*-------------------------------------------------------------------------------
* START
*-------------------------------------------------------------------------------
*/
int main( int argc, char* argv[] )
{
	int rc ;

	libusb_init( &usb_context );

	/* Parse command line */
	rc=parse_argv( argc, argv, OPTIONS_N ) ;
	if( rc )
	{
		return rc;
	}

	return rc;
}

/*
*-----------------------------------------------------------------------------
*                               enum_devices
*-----------------------------------------------------------------------------
*/

typedef int (*USB_ENUM_DEV_CALLBACK)(
										libusb_device_handle* hdev, 
										struct libusb_device_descriptor *desc,
										void *context);

int enum_devices(USB_ENUM_DEV_CALLBACK callback, void* context )
{
	int status = 0;
	int i, count;

	libusb_device** list;
	struct libusb_device_descriptor desc;
    libusb_device_handle *hdev;


	count = libusb_get_device_list( usb_context, &list );

	for(i=0;i<count;i++)
	{
		status = libusb_get_device_descriptor(list[i], &desc);

		if( status )
		{
			printf("ERROR: Cannot obtain device descriptor, status=%d\n", status );
			break ;
		}
		
		if( ( (desc.idVendor == SUB_VID) && (desc.idProduct == SUB_PID)) ||
			((desc.idVendor == MTUSB_VID) && (desc.idProduct == MTUSB_PID))	)
		{
			status = libusb_open( list[i], &hdev );
			if(status)
			{
				printf("ERROR: Cannot open USB device, status=%d\n", status );
				break ;
			}

			if( 0 == callback( hdev, &desc, context ) )
			{
				break ;
			}

			libusb_close( hdev );
		}
	}

	libusb_free_device_list( list, 1 );

	return status ;
}


/*
*-----------------------------------------------------------------------------
*                               display_device_info
*-----------------------------------------------------------------------------
*/
int display_device_info(	libusb_device_handle* hdev, 
							struct libusb_device_descriptor *desc)
{
	int status;
	int bytes, mode;
	char *model;
	char buf[16];
	const struct sub_version* version;

	bytes = sub_get_serial_number( hdev, buf, sizeof(buf));
	if(bytes < 0)
	{
		printf("ERROR: Cannot obtain serial number, status=%d\n", bytes );
		return bytes;
	}
	buf[bytes]=0;

	model = desc->idVendor == SUB_VID ? "SUB20" : "MTUSB";

	version = sub_get_version( hdev );

	status = sub_get_mode( hdev, &mode ) ;

	if( status )
	{
		printf("ERROR: Cannot obtain current mode from device, %d\n", status ) ;
		return status ;
	}

	if( mode == SUB_BOOT_MODE )
	{
		printf("%s S/N %s Boot version %d.%d\n", 
			model, buf, version->sub_device.boot >> 4, 
			version->sub_device.boot & 0x0f);
	}
	else
	{
		printf("%s S/N %s FW version %d.%d.%d Boot version %d.%d\n", 
			model, buf, version->sub_device.major, 
			version->sub_device.minor, version->sub_device.micro,
			version->sub_device.boot >> 4, version->sub_device.boot & 0x0f);
	}

	return 0;
}

/*
*-----------------------------------------------------------------------------
*                               list_usb_devices_callback
*-----------------------------------------------------------------------------
*/
int list_usb_devices_callback(	libusb_device_handle* hdev,
								struct libusb_device_descriptor *desc,
								void *context)
{
	return (display_device_info(hdev, desc) == 0 ); /* continue if success*/
}

/*
*-----------------------------------------------------------------------------
*                               count_usb_devices_callback
*-----------------------------------------------------------------------------
*/
int count_usb_devices_callback(	libusb_device_handle* hdev,
								struct libusb_device_descriptor *desc,
								void *context)
{
	(*(int*)context)++;
	return 1; /* continue */
}


/*
*-----------------------------------------------------------------------------
*                               process_arg
*-----------------------------------------------------------------------------
* Dscr: process argument  
*-----------------------------------------------------------------------------
* Input : option index, equ_ptr - pointer to option value
* Output: 0 - continue
*         <0 - error code to exit 
*         >0 - code to exit
* Notes : call exit(1) in case of error in argument
*-----------------------------------------------------------------------------
*/
int process_arg( int opt_index, char* equ_ptr )
{
	switch( usage_descriptor[ opt_index ].case_code )
	{
    case HELP_CMD_CASE:
        usage( 1, OPTIONS_N );
        return 1;
	case VER_CMD_CASE:
        XPRINT_VERSION;
        break;
    case LIST_CMD_CASE:
		enum_devices( list_usb_devices_callback, NULL );
        break;
	case UPD_CMD_CASE:
		update_firmware(equ_ptr);
		break;
	case NON_INT_CMD_CASE:
		do_not_ask=1;
		break;
	}

	return 0;
}

/*
*-----------------------------------------------------------------------------
*                               do_update_firmware
*-----------------------------------------------------------------------------
*/
int do_update_firmware(libusb_device_handle* hdev, 
					   upd_fw_context_t *upd_fw_context)
{
	int status = 0 ;
	int write_size, size ;
	char *ptr=upd_fw_context->buf;
	size = upd_fw_context->size ;

	while( size )
	{
		write_size = min( size, WRITE_BUFFER_SIZE );
		status = sub_update_nvram( hdev, ptr, write_size );
		if( status )
		{
			printf("ERROR: Updating firmware failed, status=%d\n", status );
			break ;
		}

		ptr += write_size ;
		size -= write_size ;

		printf("Updating ... %d %%\r", 
			(upd_fw_context->size - size) * 100 / upd_fw_context->size );
		fflush(stdout);

	}

	printf("\n");

	return status ;
}


/*
*-----------------------------------------------------------------------------
*                               find_device_callback
*-----------------------------------------------------------------------------
*/
int find_device_callback(	libusb_device_handle* hdev,
							struct libusb_device_descriptor *desc,
							void *context)
{
	int bytes;
	char buf[128];
	uint16_t sn;
	find_device_context_t *find_device_context=(find_device_context_t *)context;
	bytes = sub_get_serial_number( hdev, buf, sizeof(buf));
	if(bytes < 0)
	{
		printf("ERROR: Cannot obtain serial number, status=%d\n", bytes );
		find_device_context->upd_fw_context->status=bytes;
		return 0; /* stop */
	}
	buf[bytes]=0;
	sn=strtoul(buf, NULL, 16);
	if( find_device_context->sn == sn )
	{
		printf("\n" ) ;
		find_device_context->upd_fw_context->status = do_update_firmware( hdev, 
			find_device_context->upd_fw_context );
		find_device_context->upd_fw_context->done = 1 ;

		return 0;	/* stop */
	}

	return 1; /* continue */
}


/*
*-----------------------------------------------------------------------------
*                               update_firmware_callback
*-----------------------------------------------------------------------------
*/

int update_firmware_callback(	libusb_device_handle* hdev,
								struct libusb_device_descriptor *desc,
								void *context)
{
	char buf[128];
	int mode, bytes ;
	int i;

	upd_fw_context_t *upd_fw_context = (upd_fw_context_t *)context ;

	find_device_context_t find_device_context;

	upd_fw_context->status = display_device_info(hdev, desc) ;
	if( upd_fw_context->status )
	{
		return 0 ; /* stop if not success*/
	}

	if( 0 == do_not_ask )
	{
		for(;;)
		{
			printf("Do you want to update FW Y/[N]? ");
			if( !fgets( buf, sizeof(buf), stdin ) )
                return 0;
			if( 'y' == *buf || 'Y' == *buf || 
				'n' == *buf || 'N' == *buf  || 
				'\n' == *buf )
			{
				break;
			}
		}

		if( 'N' == *buf || 'n' == *buf )
		{
			return 1 ; /* next device */
		}
	}

	upd_fw_context->status = sub_get_mode( hdev, &mode ) ;

	if( upd_fw_context->status )
	{
		printf("ERROR: Cannot obtain current mode from device, %d\n", upd_fw_context->status ) ;
		return 0; /* stop */
	}

	bytes = sub_get_serial_number( hdev, buf, sizeof(buf));
	if(bytes < 0)
	{
		printf("ERROR: Cannot obtain serial number, status=%d\n", bytes );
		return 0; /* stop */
	}

	buf[bytes] = 0 ;

	upd_fw_context->done = 0 ;

	if( mode != SUB_BOOT_MODE )
	{
		printf("Set boot mode ... " ) ;
		upd_fw_context->status = sub_set_mode( hdev, SUB_BOOT_MODE ) ;

		if( upd_fw_context->status )
		{
			printf("\n" ) ;
			printf("ERROR: Cannot set BOOT mode, %d\n", upd_fw_context->status ) ;
			return 0; /* stop */
		}

		printf("OK\n") ;

		find_device_context.vid = desc->idVendor ;
		find_device_context.pid = desc->idProduct ;
		find_device_context.sn = strtoul(buf, NULL, 16);
		find_device_context.upd_fw_context = upd_fw_context;

		/* wait for the same device to appear again, but in BOOT mode now */
		for(i=0;i<WAIT_FOR_DEVICE_TIMEOUT;i++)
		{
			printf("Waiting for %s S/N %s ... %d\r", 
				SUB_VID == desc->idVendor ? "SUB20" : "MTUSB", 
				buf, WAIT_FOR_DEVICE_TIMEOUT - i - 1 ) ;

			fflush(stdout);
			sleep( 1 );

			upd_fw_context->status = enum_devices( find_device_callback, 
				&find_device_context );

			if( upd_fw_context->status )
			{
				printf("ERROR: Device access failed, %d\n", upd_fw_context->status ) ;
				return 0; /* stop */
			}

			if( find_device_context.upd_fw_context->done )
			{
				break;
			}
		}

		if( !find_device_context.upd_fw_context->done )
		{
			printf("ERROR: Cannot find device\n" ) ;
			return 0; /* stop */
		}
	}
	else
	{
		upd_fw_context->status = do_update_firmware( hdev, upd_fw_context );

		if( upd_fw_context->status )
		{
			return 0; /* stop */
		}
	}

	return 1; /* continue */
}

/*
*-----------------------------------------------------------------------------
*                               count_usb_devices_callback
*-----------------------------------------------------------------------------
*/
int update_firmware(char *filename)
{
	int status;
	int count;
	FILE *stream=NULL ;
	upd_fw_context_t upd_fw_context;

	do
	{
		memset( &upd_fw_context, 0, sizeof(upd_fw_context));

		count = 0;
		status = enum_devices( count_usb_devices_callback, &count );
		if(status)
		{
			printf("ERROR: Cannot count devices, status=%d\n", status );
			break;
		}

		if( count == 0 )
		{
			printf("ERROR: No device found\n");
			status=-1;
			break;
		}

		stream = fopen(filename, "rb" );

		if( !stream )
		{
			printf("ERROR: Cannot open image file %s\n", filename );
			break ;
		}

		fseek(stream, 0, SEEK_END);
		upd_fw_context.size = ftell( stream ) ;
		fseek(stream, 0, SEEK_SET);

		upd_fw_context.buf = (char*) malloc( upd_fw_context.size ) ;
		if(!upd_fw_context.buf)
		{
			printf("ERROR: Memory allocation failed\n");
			break ;
		}

		if( upd_fw_context.size != 
			fread(upd_fw_context.buf, 1, upd_fw_context.size, stream) )
		{
			printf("ERROR: Image file read failed\n" );
			break ;
		}

		status = enum_devices( update_firmware_callback, &upd_fw_context );

		if( 0 == upd_fw_context.status )
		{
			printf("Please re-plug device(s) to exit from the BOOT mode\n" );
		}

	}while(0);

	if( stream )
	{
		fclose(stream);
	}

	if( upd_fw_context.buf )
	{
		free(upd_fw_context.buf);
	}

	return status;
}
